#include <time.h>
#include "log.h"
#include "scenegraph.h"
#include "font.h"

SceneGraph* GlobalGraph = 0;

SceneGraph::SceneGraph()
{
	Screen = 0;
	Initialized = 0;
	SDLFlags = SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_OPENGL;
	BPP = 0;	// aka current bytes per pixel
	Resolution.x = 640;
	Resolution.y = 480;
	TimeScale = 1.0f;
	TimeActive = 1;
	MouseButtons[0] = 0;
	MouseButtons[1] = 0;
	MouseButtons[2] = 0;
	MouseButtons[3] = 0;
	InitLog("log out.txt");
}

SceneGraph::~SceneGraph()
{
	ShutdownLog();
}

// changes time scale
void SceneGraph::SetTimeScale( float scale)
{
	if (scale <= 0)
		TimeScale = 0;
	else
		TimeScale = scale;
}

void SceneGraph::StopTimer()
{
	TimeActive = 0;
}

void SceneGraph::StartTimer()
{
	TimeActive = 1;
}


void InitSceneGraph(int width, int height, int fullscreen)
{
	GlobalGraph = new SceneGraph();
	GlobalGraph->Initialize(width, height, fullscreen);
}

void ShutdownSceneGraph()
{
	GlobalGraph->Shutdown();
	delete GlobalGraph;
}

SceneGraph* GetSceneGraph() { return GlobalGraph; }

void SceneGraph::AddNote( char* text, float displaytime)
{
	NoteInfo toadd;
	toadd.DisplayTime = displaytime;
	toadd.Text = (std::string)text;
	Notes.push_back(toadd);
}


Point_2d SceneGraph::GetResolution()
{
	return Resolution;
}


int SceneGraph::Initialize(int width, int height, int fullscreen)
{
	LogMessageX("initializing display");
	Resolution.x = width;
	Resolution.y = height;
	int error;
	if (fullscreen)
		SDLFlags |= SDL_FULLSCREEN;
	error = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER);

	if (error == -1)
	{
		LogMessageX("Could not initialize SDL");
		return 0;
	}

	Screen = SDL_SetVideoMode(Resolution.x, Resolution.y, BPP, SDLFlags);
    if ( Screen == 0 )
	{
		LogMessageX("could not initialize video");
		return 0;
    }

	glViewport(0,0,(GLsizei)(Resolution.x),(GLsizei)(Resolution.y));						
	glMatrixMode(GL_PROJECTION);									
	glLoadIdentity();												
	glOrtho(0, (GLfloat)(Resolution.x), 0, (GLfloat)(Resolution.y), -10.0f, 10.0f);
	glMatrixMode(GL_MODELVIEW);									
	glLoadIdentity();											

	glEnable(GL_TEXTURE_2D);
	glClearColor(0.0f,0.0f,0.0f,0.5f);							
	glClearDepth(1.0f);											
//	glDepthFunc(GL_LEQUAL);										
	glEnable(GL_BLEND);
//	glEnable(GL_DEPTH_TEST);
	glDisable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);									
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	
	LogMessageX("set viewport/enabled blending, depthtest, smoothshading");
	
	SDL_GetMouseState(&MouseX, &MouseY);
	MouseY = Resolution.y - MouseY;

	SDL_EnableUNICODE(1);	// enable keyboard translation
	SDL_WM_SetCaption("Charged Particle Simulator", "Particle Sim");
	LogMessageX("setting window caption");
	srand(clock());	// seed the random number generator for later use

	Initialized = 1;
	return 1;
}


void SceneGraph::Shutdown()
{
	LogMessageX("shutting down display");
	for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
		(*fi)->Shutdown();

	Layers.resize(0);
	SDL_Quit();
}


void SceneGraph::AddLayer(SceneLayer* layer, char* name, int order)
{
	layer->Name = name;
	layer->Precedence = order;

	// if empty just insert
	if (Layers.size() == 0)
	{
		Layers.push_back(layer);
		return;
	}


	// find one with higher precedence and insert layer before
	for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
	{
		if ((*fi)->Precedence >= order)
		{
			Layers.insert(fi, 1,layer);
			return;
		}
	}

	// this layer has highest order
	Layers.push_back(layer);
	return;


}

SceneLayer* SceneGraph::GetLayer( char* name)
{
	for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
	{
		if ((*fi)->Name.find(name) != -1)
			return *fi;
	}
	return 0;
}


void SceneGraph::RemoveLayer(SceneLayer* layer)
{
	// add pending removal
	ToRemove.push_back(layer);
	
}

void SceneGraph::QuitApp()
{
	Active = 0;
}

void SceneGraph::PumpMessages()
{
	SDL_Event event; 

	while(SDL_PollEvent(&event))
	{  
		switch(event.type)
		{  
		case SDL_VIDEORESIZE:
			{
				Resolution.x = event.resize.w;
				Resolution.y = event.resize.h;
				glViewport(0,0,(GLsizei)(Resolution.x),(GLsizei)(Resolution.y));						
				glMatrixMode(GL_PROJECTION);									
				glLoadIdentity();												
				glOrtho(0, (GLfloat)(Resolution.x), 0, (GLfloat)(Resolution.y), -10.0f, 10.0f);
				glMatrixMode(GL_MODELVIEW);									
				glLoadIdentity();	
			}
			break;
		case SDL_QUIT:
			{
				LogMessageX("recieved SDL_QUIT message");
				QuitApp();
			}
			break;
		case SDL_KEYDOWN:   
			{
				if (event.key.keysym.sym == SDLK_ESCAPE)
				{
					if (event.key.state == 1)
					{
						// do a double check seeing as how it crashed
						// by making false assumption of keydown
						if (GetKeyState(VK_ESCAPE) & 128)
						{
							QuitApp();
						}
					}
				}
				// do some translation
				// apparently sdl does not translate numpad numerals
				// to their ascii equivalent
				if ((event.key.keysym.sym >= SDLK_KP0) && (event.key.keysym.sym <= SDLK_KP9) && (event.key.keysym.mod & KMOD_NUM))
				{
					int difference = '0' - SDLK_KP0;
					event.key.keysym.unicode = event.key.keysym.sym + difference;
				}

				for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
					(*fi)->OnKey(&event.key);
		
			}	
			break;
		case SDL_MOUSEMOTION:
			{
	
				// program follows +y up
				MouseX = event.motion.x;
				MouseY = Resolution.y - event.motion.y;

				
				for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
				{
					if ((*fi)->GetActive())
						(*fi)->OnMouse(MouseX,MouseY,MouseButtons,-1);
				}
			}
			break;
		case SDL_MOUSEBUTTONDOWN:
		case SDL_MOUSEBUTTONUP:
			{
				
				MouseButtons[event.button.button-1] = event.button.state;
				for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
				{
					if ((*fi)->GetActive())
						(*fi)->OnMouse(MouseX,MouseY,MouseButtons,event.button.button-1);
				}
			}
			break;

		}
		// if any items told to remove selves, do so
		ClearRemove();
	}

}

void SceneGraph::ClearNotes()
{
	Notes.resize(0);
}

void SceneGraph::ClearRemove()	// iterate through toremove and clear
{
	if (ToRemove.size())
	{
		for (std::list<SceneLayer*>::iterator i = ToRemove.begin(); i != ToRemove.end(); i++)
		{
			Layers.remove( *i);
		}
		ToRemove.resize(0);
	}
	
}

int SceneGraph::Run()
{
	if (!Initialized)
	{
		LogMessageX("not initialized, cannot Run()");
		return 0;
	}
	// first pump and send messages
	LogMessageX("pumping messages");
	PumpMessages();

	static float lasttime = (float)SDL_GetTicks()/1000.0f;
	float time = (float)SDL_GetTicks()/(float)1000.0f;
	float delta = time - lasttime;
	lasttime = time;
	
	if (TimeActive)
	{
		

	
		float framedelta = delta*TimeScale;
		
		// Run per frame Logic
		for (std::list<SceneLayer*>::iterator fi = Layers.begin(); fi != Layers.end(); fi++)
		{
			if ((*fi)->GetActive())
				(*fi)->Tick(framedelta);
		}
	}

	LogMessageX("clearing gl window");
	glClear (GL_COLOR_BUFFER_BIT);		
	glLoadIdentity();											
	glTranslatef(0, 0, -1.0f);	

	for (std::list<SceneLayer*>::iterator i = Layers.begin(); i != Layers.end(); i++)
	{
		if ((*i)->GetActive())
			(*i)->Draw();
	}

	// draw any notes that need to be draw
	// 
	float text_x, text_y;
	text_x = 50;
	text_y = Resolution.y - 170;
	std::list<NoteInfo>::iterator n = Notes.begin();
	
	while( n != Notes.end())
	{
		n->DisplayTime -= delta;

		if (n->DisplayTime <= 0)
		{
			// remove n
			std::list<NoteInfo>::iterator old = n;
			n++;
			Notes.erase(old);
		}
		else
		{
			Font* f = GetFont();

			f->Draw( text_x, text_y, (char*)n->Text.c_str(), n->Text.size());
			text_y += f->GetSize((char*)n->Text.c_str(), n->Text.size()).y;
			text_y -= f->GetLineHeight();
			
			n++;
		}
	}

	LogMessageX("swapping buffers using SDL_GL_SwapBuffers");
	SDL_GL_SwapBuffers();
	
	ClearRemove(); // clear any items that requested removal
	return Active;
}
